/* bttv.c - functions for handling v4l and lowlevel bttv input
 *
 * Parts Copyright (C) Ewald Snel 1998-1999
 * Tuner/Channel stuff from MPlayer v4l interface (C) Alex Beregszaszi
 * The rest: Copyright (C) 2003 Rik Snel <snel@phys.uu.nl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; only version 2 of the License, not
 * any other version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 * $Id: bttv.c,v 1.9 2004/02/17 11:06:27 rsnel Exp $
 */
#if 0
/* this definition is problematic for some, so
   we implement round() ourselves */
#define _ISOC99_SOURCE /* to get round() from math.h */
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <sys/io.h>
#include <sys/time.h>
#include <fcntl.h>
#include <unistd.h>
#include <linux/videodev.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <assert.h>
#include "cc2y4m.h"
#include "bttv.h" 

/* Assuming that the capture buffer size is 2129920:
 *
 * A Raw8X frame in TEST mode not all pixels are used; each capture buffer
 * contains: (this data probably depends of the doings of the *Scaler())
 * - 655360 (=10*2^16) pixels of almost continuesly sampled data (288 lines of
 *   2268 pixels and one of 2176 pixels) This data should be treated as the
 *   first field of the captured frame.
 * - 409600 (=100*2^12) unused pixels
 * - 655360 (=10*2^16) pixels of almost continuously sampled data This data
 *   should be treated as the second field 
 * - 409600 (=100*2^12) unused pixels
 *
 * Counting from zero, the pixels 2014, 2015, ..., 2100 are not sampled
 * correctly. the are all too dark, and pixels 2045-2070 are all equal to pixel
 * 2044. This is probalby due to a periodic (every line) reinitiazation of the
 * ADC.
 * 
 * The appearance of the reset with an interval of 2268 pixels tells us that
 * the internal line length really is 2268, otherwise the line length could
 * just as easily be 2270 or 2265 (by a change in the sampling freqency of the
 * ADC). The periodic reset forces us to use lines with a length of 2268
 * pixels. (otherwise we would those reset-artefacts in our active picture or
 * color burst)
 */

/* code from ccd.c, distributed with cabletv-1.3.9,
 * (probably) Copyright (C) Ewald Snel (1998-1999) */
static void doScale(bttv_t *bttv)
{
	local_regs_t *r = bttv->regs;
	short s;
	int field_width = (r->o_hactive_msb << 8) | r->o_hactive_lo;
	int vactive = (r->o_vactive_msb << 8) | r->o_vactive_lo;

	if (bttv->hactive == 0)
		s = 0;
	else
		s =  (short) ((2L*bttv->hdelay*field_width / 
					bttv->hactive + 1) >> 1);

	VERBOSE("doScale, hdelay=%d\n", s);
	r->e_hdelay_msb = r->o_hdelay_msb = (s >> 8) & 3;
	r->e_hdelay_lo = r->o_hdelay_lo = s & 0xff;

	if (field_width == 0)
		s = 0;
	else
		s = (short) ((8192L*bttv->hactive / field_width - 8192L + 1) 
				>> 1);

	VERBOSE("doScale, hscale=%d\n", s);
	r->e_hscale_hi = r->o_hscale_hi = (s >> 8) & 0xff;
	r->e_hscale_lo = r->o_hscale_lo = s & 0xff;

	if (bttv->field_height == (vactive >> 1) || bttv->field_height == -2)
		s = 0;
	else
		s = (short) ((0x10000L + 512L - 
					((512L*vactive /bttv->field_height + 1) 
					>> 1)) & 0x1FFFL);

	VERBOSE("doScale, vscale=%d\n", s);
	r->e_vscale_hi = r->o_vscale_hi = (s >> 8) & 0x1f;
	r->e_vscale_lo = r->o_vscale_lo = s & 0xff;
}

static void scalerInit(bttv_t *bttv)
{
	local_regs_t *r = bttv->regs;
	int f, a;

	f = (r->o_hscale_hi << 8) | r->o_hscale_lo;
	bttv->hdelay = ((f + 4096) * ((r->o_hdelay_msb << 8) | 
				r->o_hdelay_lo) + 2048) >> 12;
	bttv->hactive = ((f + 4096) * ((r->o_hactive_msb << 8) | 
				r->o_hactive_lo) + 2048) >> 12;
	f = (r->o_vscale_hi << 8) | r->o_vscale_lo;
	a = (r->o_vactive_msb << 8) | r->o_vactive_lo;
	bttv->field_height = (f == 0) ? a : (512L*a / (0x2200L - f) - 2);
	VERBOSE("hdelay=%d, hactive=%d, field_height=%d\n",
			bttv->hdelay, bttv->hactive, bttv->field_height);
	doScale(bttv);
}
/* end of code from ccd.c */

#define XT0FRQ 28636363

/* always assume pre devisor = 2, post divisor = 6 */
double bttv_set_pll(bttv_t *bttv, double freq) {
	local_regs_t *r = bttv->regs;
	double times = 12*65536*freq/XT0FRQ;
	int ipart;
/*	int mult = (int)round(times); */
	int mult = (int)times;
        if (times - (double) mult >= 0.5) mult++;
	r->pll_x = 1;
	r->pll_c = 0;
	r->pll_f_lo = mult&0xFF;
	r->pll_f_hi = (mult>>8)&0xFF;
	ipart = mult>>16;
	assert(ipart >= 6 && ipart < 64);
	r->pll_i = ipart;
	assert(times > 0);
#if 0
	DEBUG("setting PLL to %lf (err=%lf)\n", XT0FRQ*(double)mult/(12*65536),
			XT0FRQ*(double)mult/(12*65536) - freq);
#endif
	return XT0FRQ*(double)mult/(12*65536) - freq;
}

double bttv_get_pll(bttv_t *bttv) {
	local_regs_t *r = bttv->regs;
	double mult = (r->pll_f_lo+256*r->pll_f_hi)/65536. + r->pll_i;
	double div = ((r->pll_c == 0) ? 6 : 4)*((r->pll_x == 0) ? 1 : 2);
	return XT0FRQ*mult/div;
}

void bttv_pll_up(bttv_t *bttv) {
	local_regs_t *r = bttv->regs;
	int pll_f_hi = r->pll_f_hi;
	int pll_f_lo = r->pll_f_lo;
	int pll_i = r->pll_i;
	if (pll_f_lo == 255) {
		r->pll_f_lo = 0;
		if (pll_f_hi == 255) {
			r->pll_f_hi = 0;
			if (pll_i >= 64) assert(0);
			else r->pll_i = pll_i + 1;
		} else r->pll_f_hi = pll_f_hi + 1;
	} else r->pll_f_lo = pll_f_lo + 1;
}
		
void bttv_schedule_down(bttv_t *bttv) {
	pthread_mutex_lock(&bttv->cmutex);
	bttv->updown = -1;
	pthread_mutex_unlock(&bttv->cmutex);
}
		
void bttv_schedule_updown(bttv_t *bttv, int updown) {
	pthread_mutex_lock(&bttv->cmutex);
	bttv->updown = updown;
	pthread_mutex_unlock(&bttv->cmutex);
}

void bttv_schedule_pll(bttv_t *bttv, double newfrq) {
	pthread_mutex_lock(&bttv->cmutex);
	bttv->newfrq = newfrq;
	pthread_mutex_unlock(&bttv->cmutex);
}

void bttv_set_vtotal(bttv_t *bttv, int vtotal) {
	local_regs_t *r = bttv->regs;
	vtotal--;
	r->vtotal_hi = (vtotal>>8)&0xff;
	//DEBUG("setting vtotal to %d\n", vtotal + 1);
	r->vtotal_lo = vtotal&0xff;
}

void bttv_pll_down(bttv_t *bttv) {
	local_regs_t *r = bttv->regs;
	int pll_f_hi = r->pll_f_hi;
	int pll_f_lo = r->pll_f_lo;
	int pll_i = r->pll_i;
	if (pll_f_lo == 0) {
		r->pll_f_lo = 255;
		if (pll_f_hi == 0) {
			r->pll_f_hi = 255;
			if (pll_i <= 6) assert(0);
			else r->pll_i = pll_i - 1;
		} else r->pll_f_hi = pll_f_hi - 1;
	} else r->pll_f_lo = pll_f_lo - 1;
}
		
int bttv_get_vtotal(bttv_t *bttv) {
	local_regs_t *r = bttv->regs;
	return (r->vtotal_lo|(r->vtotal_hi<<8)) + 1;
}

int bttv_init(bttv_t *bttv, char *dev, unsigned long chan) {
	int i;
	unsigned long int device, addr = 0, pcireg;
	struct video_tuner tuner;
	local_regs_t *r;
	assert(bttv && dev);

	/* defaults */
	bttv->fd = -1;
	bttv->memfd = -1;
	bttv->map = (char*)-1;
	bttv->regs = (local_regs_t*)-1;

	if (geteuid() != 0) {
		ERROR("we need to access local registers of the Bt8x[89] chip; root privileges are needed\n");
		return -1;
	}

	if ((bttv->fd = open(dev, O_RDWR)) == -1) {
		ERROR("opening %s: %s\n", dev, strerror(errno));
		return -1;
	}
	VERBOSE("%s successfully openend\n", dev);

	if (ioctl(bttv->fd, VIDIOCGCAP, &bttv->vc)) {
		ERROR("getting video capabilities, not v4l device?\n");
		close(bttv->fd);
		return -1;
	}

	/* aquire info on grab buffers... */
	if (ioctl(bttv->fd, VIDIOCGMBUF, &bttv->buf_info) == -1) {
		ERROR("VIDIOCGMBUF, bailing out\n");
		close(bttv->fd);
		return -1;
	}

	/* check number and size */
	if (bttv->buf_info.frames < 4) {
		ERROR("not enough frames in provided capture buffer\n");
		close(bttv->fd);
		return -1;
	}

	if (bttv->buf_info.size/bttv->buf_info.frames != 2129920) {
		ERROR("buffers have unexpected size, don't know how to do interlacing, bailing out\n");
		close(bttv->fd);
		return -1;
	}

	/* and map them */
	bttv->map = mmap(0, bttv->buf_info.size, PROT_READ|PROT_WRITE,
			MAP_SHARED, bttv->fd, 0);
	if ((int)bttv->map == -1) {
		ERROR("mapping %d capture buffers of size %d\n", 
				bttv->buf_info.frames, bttv->buf_info.size);
		return -1;
	} else VERBOSE("successfully mapped %d buffers of total size %d\n", 
			bttv->buf_info.frames, bttv->buf_info.size);
	
	for (i = 0; i < 4; i++) {
		bttv->maps[i].format = VIDEO_PALETTE_RAW;
		bttv->maps[i].frame = i;
		/* width and height seem to be ignored in raw capture more, 
		 * but values must be 'reasonable' (nonzero etc) */
		bttv->maps[i].width = 826;
		bttv->maps[i].height = 288;
	}

	/* try to find local io space of bttv chip */
	if (iopl(3) < 0) {
		ERROR("iopl(3): %s\n", strerror(errno));
		bttv_uninit(bttv);
		return -1;
	}
	
	i = 0;
	do {
		pcireg = 0x80000000 | (i << 8);

		outl(pcireg, 0xCF8);
		device = inl(0xCF8 + 4);

		switch (device) {
			case BT848:
			case BT849:
			case BT878:
			case BT879:
				bttv->type = device;
				outl(pcireg + 0x10, 0xCF8);
				addr =  inl(0xCF8 + 4)&0xfffffff0;
		}
	} while (++i < 0x10000 && addr == 0);

	if (addr == 0) {
		ERROR("unable to find BTTV chip, is your V4L device a BT8[47][89](A)?\n");
		bttv_uninit(bttv);
		return -1;
	} else VERBOSE("io space of chip is at 0x%lx\n", addr);

	switch (bttv->type) {
		case BT848:
		case BT849:
			ERROR("chip seems to be a BT84[89], unsupported at the moment\n");
			iopl(0);
			bttv_uninit(bttv);
			return -1;
			break;
		case BT878:
		case BT879:
			VERBOSE("chip seems to be a BT87%c\n", 
					bttv->type == BT878 ? '8' : '8');
			break;
	}
	
	if (iopl(0) < 0) {
		ERROR("iopl(0): %s\n", strerror(errno));
		bttv_uninit(bttv);
		return -1;
	}

	/* map the registers */
	if ((bttv->memfd = open("/dev/mem", O_RDWR)) < 0) {
		ERROR("unable to open /dev/mem: %s\n", strerror(errno));
		bttv_uninit(bttv);
		return -1;
	}
	
	bttv->regs = mmap(0, 0x1000, PROT_READ|PROT_WRITE, MAP_SHARED,
			bttv->memfd, addr);

	if ((int)bttv->regs == -1) {
		ERROR("unable to mmap local registers: %s\n", strerror(errno));
		bttv_uninit(bttv);
		return -1;
	} else VERBOSE("successfully mapped local registers\n");
	
	r = bttv->regs;

	/* setting up internal registers */
	scalerInit(bttv);
	r->xtsel = 1; 		/* select crystal 0, 28.636363MHz */
	bttv_set_pll(bttv, 35437500);
	r->tgcki = 1; 		/* enable pll */
	r->dec_fieldalign = 1;	/* align fields */
	r->dec_rate = 0;	/* don't drop frames */
	bttv_set_vtotal(bttv, 625);
	r->format = 1;		/* setting format */
	r->color_even = r->color_odd = 14; 
				/* setting Raw8X mode */
	r->test = 239;		/* don't try to sync */
	r->ext_frame = 1;	/* extend capture size */
	r->fifo_enable = 1;
	r->capture_even = r->capture_odd = 1;
	bttv->synced = -1; 
	bttv->queued0 = -1; 
	bttv->queued1 = -1; 
	bttv->exported = -1;
	bttv->avail = -1;
	pthread_mutex_init(&bttv->bmutex, NULL);
	pthread_mutex_init(&bttv->cmutex, NULL);
	pthread_cond_init(&bttv->acond, NULL);
	bttv->updown = 0;
	bttv->newfrq = 35437500;
	
	/* channel select, some of it copied from tvi_v4l in MPlayer,
	 * original code written and (C) by Alex Beregszaszi */
	bttv->channels = calloc(1, sizeof(*bttv->channels)*bttv->vc.channels);
	if (!bttv->channels) {
		ERROR("calloc failure: %s\n", strerror(errno));
		bttv_uninit(bttv);
		return -1;
	} 

	for (i = 0; i < bttv->vc.channels; i++) {
		struct video_channel *vch = bttv->channels + i;
		vch->channel = i;
		if (ioctl(bttv->fd, VIDIOCGCHAN, vch) == -1) {
			ERROR("get channel failed: %s\n", strerror(errno));
			bttv_uninit(bttv);
			return -1;
		} 
		
		VERBOSE(" %d: %s: %s%s%s%s (tuner:%d, norm:%d)\n", i,
				vch->name,
				(vch->flags&VIDEO_VC_TUNER) ?  "[tuner]" : "",
				(vch->flags&VIDEO_VC_AUDIO) ?  "[audio]" : "",
				(vch->flags&VIDEO_TYPE_TV) ?  "[tv]" : "",
				(vch->flags&VIDEO_TYPE_CAMERA) ? "[cmera]" : "",
				vch->tuners, vch->norm);
	}
	
	if (ioctl(bttv->fd, VIDIOCSCHAN, &bttv->channels[(chan == 0)]) 
			== -1) {
		ERROR("set tv input failed... bad!\n");
		bttv_uninit(bttv);
		return -1;
	}

	tuner.tuner = bttv->channels[0].tuners;

	if (chan == 0) return 0;

	if (ioctl(bttv->fd, VIDIOCGTUNER, &tuner) == -1) {
		ERROR("error getting info from tuner: %s\n", strerror(errno));
		bttv_uninit(bttv);
		return -1;
	}

	if (!(tuner.flags&VIDEO_TUNER_LOW)) chan /= 1000;

	if (ioctl(bttv->fd, VIDIOCSFREQ, &chan) == -1) {
		ERROR("set tuner freq failed: %s\n", strerror(errno));
		bttv_uninit(bttv);
		return -1;
	}
	
	return 0;
}

static void queue(bttv_t *bttv, int i) {
	//DEBUG("queueing buffer %d for capture\n", i);
	if (ioctl(bttv->fd, VIDIOCMCAPTURE, &bttv->maps[i]) == -1) {
		ERROR("queueing buffer for capture\n");
	}
}

static void my_sync(bttv_t *bttv, int i) {
	//DEBUG("waiting for buffer %d to be full\n", i);
	do errno = 0;
	while (ioctl(bttv->fd, VIDIOCSYNC, &bttv->maps[i].frame) < 0 &&
			(errno == EAGAIN || errno == EINTR)); 
	if (errno != 0) ERROR("my_sync failed: %s\n", strerror(errno));
}

void bttv_capture_thread(void *data) {
	bttv_t *bttv = data;
	char *frame;
	int frameno = 0;
	int tmp;
	DEBUG("starting capture thread\n");
	
	pthread_mutex_lock(&bttv->bmutex);
	bttv->synced = -1;
	bttv->avail = 2;
	bttv->exported = 3;
	pthread_mutex_unlock(&bttv->bmutex);

	bttv->queued0 = 0;
	bttv->queued1 = 1;
	bttv->finfos[frameno%10].pll = bttv_get_pll(bttv);
	bttv->finfos[frameno%10].vtotal = bttv_get_vtotal(bttv);
	bttv->framenoes[bttv->queued0] = frameno++;
	queue(bttv, 0);
	queue(bttv, 1);
	while (1) {
		my_sync(bttv, bttv->queued0);
		pthread_mutex_lock(&bttv->cmutex);
		bttv_set_vtotal(bttv, 625 + bttv->updown);
		bttv->finfos[frameno%10].vtotal = 625 + bttv->updown; 
		bttv->updown = 0;
		bttv->finfos[frameno%10].pll = 
			bttv->newfrq + bttv_set_pll(bttv, bttv->newfrq);
		pthread_mutex_unlock(&bttv->cmutex);
		pthread_mutex_lock(&bttv->bmutex);
		bttv->framenoes[bttv->queued1] = frameno++;
		if (bttv->avail >= 0 && bttv->synced < 0) {
			//VERBOSE("normal\n");
			bttv->synced = bttv->queued0;
			bttv->queued0 = bttv->queued1;
			bttv->queued1 = bttv->avail;
			bttv->avail = -1;
		} else if (bttv->avail < 0 && bttv->synced >= 0) {
			VERBOSE("input framedrop\n");
			tmp = bttv->synced;
			bttv->synced = bttv->queued0;
			bttv->queued0 = bttv->queued1;
			bttv->queued1 = tmp;
		} else assert(0);
		frame = bttv->map + bttv->buf_info.offsets[bttv->
			maps[bttv->synced].frame];
		pthread_cond_signal(&bttv->acond);
		pthread_mutex_unlock(&bttv->bmutex);
		queue(bttv, bttv->queued1);
	}
}

char *bttv_getnext(bttv_t *bttv, int *no) {
	char *out = NULL;
	pthread_mutex_lock(&bttv->bmutex);
	if (bttv->synced < 0) {
		/* no buffer currently available, so wait... */
		//printf("waiting for buffer\n");
	 	pthread_cond_wait(&bttv->acond, &bttv->bmutex);
	}
	bttv->avail = bttv->exported;
	bttv->exported = bttv->synced;
	bttv->synced = -1;
	*no = bttv->framenoes[bttv->exported];
	out = bttv->map + bttv->buf_info.offsets[bttv->
		maps[bttv->exported].frame];
	pthread_mutex_unlock(&bttv->bmutex);
	return out;
}
	
void bttv_uninit(bttv_t *bttv) {
	free(bttv->channels);
	if ((int)bttv->regs != -1) {
		/* FIXME: settings of bttv should be returned to normal here */
		VERBOSE("if you want to use another v4l app, be sure to reload your bttv module\n");
#if 0
		/* compiler complains about 'discarding qualifier' */
		munmap(bttv->regs, 0x1000);
		VERBOSE("unmapped local io regs\n");
#endif
	}

	if (bttv->memfd != -1) {
		close(bttv->memfd);
		VERBOSE("closed /dev/mem\n");
	}
	if ((int)bttv->map != -1) {
		munmap(bttv->map, bttv->buf_info.size);
		VERBOSE("unmapped capture buffers\n");
	}
	if (bttv->fd != -1) {
		close(bttv->fd);
		VERBOSE("closed video device\n");
	}
}

